#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Collect unstructured data from specific web page sources.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import ssl
import typing

from bs4 import BeautifulSoup
import requests
import requests_cache
import spacy

from .elem import TextChunk
from .nlp import Parser


SCRAPE_HEADERS: typing.Dict[ str, str ] = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36",  # pylint: disable=C0301
}


class Scraper:
    """
A simple HTML scraper.
    """

    def __init__ (
        self,
        config: dict,
        parser: Parser,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.parser: Parser = parser
        self.session: requests_cache.CachedSession = self.get_cache()


    def get_cache (
        self,
        ) -> requests_cache.CachedSession:
        """
Build a URL request cache session, optionally loading any
previous serialized cache from disk.
        """
        session: requests_cache.CachedSession = requests_cache.CachedSession(
            backend = requests_cache.SQLiteCache(
                self.config["scraper"]["cache_path"],
            ),
        )

        session.settings.expire_after = self.config["scraper"]["cache_expire"]

        return session


    def scrape_html (
        self,
        simple_pipe: spacy.Language,
        url: str,
        chunk_list: typing.List[ TextChunk ],
        chunk_id: int,
        ) -> int:
        """
A simple web page text scraper, which also performs chunking.
Returns the updated `chunk_id` index.
        """
        response: requests.Response = self.session.get(
            url,
            verify = ssl.CERT_NONE,
            timeout = 10,
            allow_redirects = True,
            headers = SCRAPE_HEADERS,
        )

        soup: BeautifulSoup = BeautifulSoup(
            response.text,
            features = "lxml",
        )

        scrape_doc: spacy.tokens.doc.Doc = simple_pipe("\n".join([  # pylint: disable=I1101
            para.text.strip()
            for para in soup.find_all("p")
        ]))

        chunk_id = self.parser.make_chunk(
            scrape_doc,
            url,
            chunk_list,
            chunk_id,
        )

        return chunk_id



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package definitions for Strwythura.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from .context import DomainContext

from .elem import Entity, NodeKind, StrwVocab, TextChunk

from .kg import KnowledgeGraph

from .nlp import Parser

from .opt import calc_quantile_bins, stripe_column, root_mean_square

from .profile import PerfProfiler

from .scrape import Scraper

from .strw import Strwythura, GraphRAG

from .textrank import run_textrank, cooccur_entities

from .vis import VisHTML



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the domain context, using `RDFlib` and related libraries.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from collections import defaultdict
import json
import math
import pathlib
import typing

from icecream import ic  # type: ignore
from rdflib.namespace import DCTERMS, RDF, ORG, SKOS
import gensim  # type: ignore
import lancedb  # type: ignore
import networkx as nx
import rdflib
import spacy

from .elem import Entity, NodeKind, StrwVocab, TextChunk


class DomainContext:  # pylint: disable=R0902,R0904
    """
Represent the domain context using an _ontology pipeline_ process:
vocabulary, taxonomy, thesaurus, and ontology.
    """
    STRW_PREFIX: str = "strw:"
    STRW_BASE: str = "https://github.com/DerwenAI/strwythura/#"

    SZ_PREFIX: str = "sz:"

    POS_TRANSFORM: typing.Dict[ str, str ] = {
        "PROPN": "NOUN",
    }


    def __init__ (
        self,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = {}
        self.rdf_graph: rdflib.Graph = rdflib.Graph()

        self.start_chunk_id: int = 0
        self.chunk_table: typing.Optional[ lancedb.table.LanceTable ] = None

        self.w2v_vectors: list = []
        self.w2v_model: typing.Optional[ gensim.models.Word2Vec ] = None

        self.known_lemma: typing.List[ str ] = []
        self.taxo_node: typing.Dict[ str, int ] = {}
        self.sem_layer: nx.MultiDiGraph = nx.MultiDiGraph()


    def set_config (
        self,
        config: dict,
        ) -> None:
        """
Accessor method to configure -- part of a design pattern to make the
domain context handling more "pluggable", i.e., to be subclassed and
customized for other use cases.
        """
        self.config = config

        # load the RDF-based context for the domain
        domain_path: pathlib.Path = pathlib.Path(self.config["kg"]["domain_path"])

        self.rdf_graph.parse(
            domain_path.as_posix(),
            format = "turtle",
        )


    def init_chunk_table (
        self,
        ) -> None:
        """
Initialize the chunk table in the vector store.
        """
        vect_db: lancedb.db.LanceDBConnection = lancedb.connect(self.config["vect"]["lancedb_uri"])  # pylint: disable=C0301

        self.chunk_table = vect_db.create_table(
            self.config["vect"]["chunk_table"],
            schema = TextChunk,
            mode = "overwrite",
        )

        self.start_chunk_id = 0


    def parse_lemma (
        self,
        span: list,
        *,
        debug: bool = False,
        ) -> str:
        """
Construct a parsed, lemmatized key for the given noun phrase.
        """
        lemmas: typing.List[ str ] = []

        for tok in span:
            pos: str = tok.pos_
            lemma: str = tok.lemma_.strip().lower()

            if pos in self.POS_TRANSFORM:
                pos = self.POS_TRANSFORM[pos]

            lemmas.append(f"{pos}.{lemma}")

        lemma_key: str = " ".join(lemmas)

        if debug:
            ic(lemma_key, span)

        return lemma_key


    def get_lemma_index (
        self,
        lemma_key: str,
        ) -> int:
        """
Lookup the UID for nodes in the semantic layer, based on a parsed
lemma key for a known entity.
        """
        return self.known_lemma.index(lemma_key)


    def add_lemma (
        self,
        lemma_key: str,
        ) -> bool:
        """
Add a known entity, indexed by its parsed lemma key.
        """
        prev_known: bool = True

        if lemma_key not in self.known_lemma:
            self.known_lemma.append(lemma_key)
            prev_known = False

        return prev_known


    def get_first_lemma (
        self,
        concept_iri: rdflib.term.Node,
        ) -> str:
        """
Get the primary lemma for a `skos:Concept` entity.
        """
        lemma_phrase_iri: rdflib.term.URIRef = self.rel_iri(StrwVocab.LEMMA_PHRASE)

        return next(
            self.rdf_graph.objects(concept_iri, lemma_phrase_iri)
        ).toPython()  # type: ignore


    def form_concept (
        self,
        fragment: str,
        ) -> rdflib.term.URIRef:
        """
Lookup a `skos:Concept` entity by its IRI.
        """
        iri: str = f"{self.STRW_BASE}{fragment}"
        concept_iri: rdflib.term.URIRef = rdflib.term.URIRef(iri)

        return concept_iri


    def rel_iri (
        self,
        rel: StrwVocab,
        ) -> rdflib.term.URIRef:
        """
Accessor to construct a `URIRef` for a relation within the `strw:` vocabulary.
        """
        return rdflib.term.URIRef(
            self.STRW_BASE + rel.value.replace(self.STRW_PREFIX, "")
        )


    def populate_taxo_node (
        self,
        concept_iri: rdflib.term.URIRef,
        ) -> typing.Tuple[ int, str, dict ]:
        """
Populate a semantic layer node from a `skos:Concept` entity in the taxonomy.
        """
        lemma_phrase_iri: rdflib.term.URIRef = self.rel_iri(StrwVocab.LEMMA_PHRASE)

        lemmas: typing.List[ str ] = [
            lemma.toPython()  # type: ignore
            for lemma in self.rdf_graph.objects(concept_iri, lemma_phrase_iri)
        ]

        lemma_key: str = lemmas[0]
        self.add_lemma(lemma_key)

        node_id: int = self.get_lemma_index(lemma_key)
        label: str = concept_iri.n3(self.rdf_graph.namespace_manager)
        self.taxo_node[label] = node_id

        text: str = self.rdf_graph.value(
            concept_iri,
            SKOS.definition,
        ).toPython()  # type: ignore

        iri: str = self.rdf_graph.value(
            concept_iri,
            DCTERMS.identifier,
        ).toPython()  # type: ignore

        self.chunk_table.add([  # type: ignore
            TextChunk(
                uid = self.start_chunk_id,
                url = iri,
                sent_id = 0,
                text = text,
            )
        ])

        self.start_chunk_id += 1

        self.sem_layer.add_node(
            node_id,
            kind = NodeKind.TAXONOMY.value,
            key = lemma_key,
            label = label,
            text = text,
            iri = iri,
            rank = 0.0,
            count = 0,
        )

        # scheduled as relations to get added, once the nodes are in place
        attrs = {
            "lemmas": lemmas,
            "broader": [
                self.get_first_lemma(node)
                for node in self.rdf_graph.objects(concept_iri, SKOS.broader)
            ],
            "narrower": [
                self.get_first_lemma(node)
                for node in self.rdf_graph.objects(concept_iri, SKOS.narrower)
            ],
            "related": [
                self.get_first_lemma(node)
                for node in self.rdf_graph.objects(concept_iri, SKOS.related)
            ],
        }

        return node_id, lemma_key, attrs


    def load_taxonomy (
        self,
        ) -> None:
        """
Iterate through `skos:Concept` entities, loading into `NetworkX`
        """
        node_map: typing.Dict[ str, int ] = {}
        attr_map: typing.Dict[ int, dict ] = {}

        # first pass: populate nodes for the `skos:Concept` entities
        for concept_iri in self.rdf_graph.subjects(RDF.type, SKOS.Concept):
            node_id, lemma_key, attr = self.populate_taxo_node(concept_iri)  # type: ignore
            node_map[lemma_key] = node_id
            attr_map[node_id] = attr

        # second pass: add relations
        for src_id, attr in attr_map.items():
            for rel in [ "broader", "narrower", "related" ]:
                rel_iri: str = f"skos:{rel}"

                for dst_key in attr[rel]:
                    dst_id: int = node_map[dst_key]

                    self.sem_layer.add_edge(
                        src_id,
                        dst_id,
                        key = rel_iri,
                        prob = 1.0,
                    )


    def parse_er_export (  # pylint: disable=R0912,R0913,R0914,R0915
        self,
        datasets: typing.List[ str ],
        *,
        export_path: typing.Optional[ pathlib.Path ] = None,
        rdf_path: typing.Optional[ pathlib.Path ] = None,
        language: str = "en",
        debug: bool = True,
        ) -> None:
        """
Parse the Senzing entity resolution results exported as JSON.
        """
        if export_path is None:
            export_path = pathlib.Path(self.config["er"]["export_path"])

        if rdf_path is None:
            rdf_path = pathlib.Path(self.config["er"]["thesaurus_path"])

        rdf_list: typing.List[ str ] = [
            """
@prefix strw:  <https://github.com/DerwenAI/strwythura/#> .
@prefix sz:    <https://senzing.com/#> .

@prefix dct:   <http://purl.org/dc/terms/> .
@prefix org:   <http://www.w3.org/ns/org#> .
@prefix rdf:   <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix skos:  <http://www.w3.org/2004/02/skos/core#> .
            """
        ]

        org_map: typing.Dict[ str, str ] = {}
        parent: typing.Dict[ str, str ] = {}

        simple_pipe: spacy.Language = spacy.load(self.config["nlp"]["spacy_model"])

        # load the data records
        data_records: typing.Dict[ str, dict ] = {}

        for filename in datasets:
            data_path: pathlib.Path = pathlib.Path(filename)

            with open(data_path, encoding = "utf-8") as fp:
                for line in fp:
                    rec: dict = json.loads(line)
                    record_id: str = self.SZ_PREFIX + rec["DATA_SOURCE"].replace(" ", "_").lower() + "_" + rec["RECORD_ID"]  # pylint: disable=C0301
                    data_records[record_id] = rec

        # parse the JSON export
        with open(export_path, encoding = "utf-8") as fp:
            for line in fp:
                data: dict = json.loads(line)

                entity_id: str = self.SZ_PREFIX + str(data["RESOLVED_ENTITY"]["ENTITY_ID"])
                ent_descrip: str = ""
                ent_type: str = ""

                rec_list: typing.List[ dict ] = []
                rel_list: typing.List[ dict ] = []

                for rec in data["RESOLVED_ENTITY"]["RECORDS"]:
                    ent_descrip = rec["ENTITY_DESC"]

                    record_id = rec["RECORD_ID"]
                    data_source: str = rec["DATA_SOURCE"].replace(" ", "_").lower()
                    rec_iri: str = f"{self.SZ_PREFIX}{data_source}_{record_id}"
                    parent[rec_iri] = entity_id

                    pred_iri: str = "skos:exactMatch"

                    rec_list.append({
                        "pred": pred_iri,
                        "obj": rec_iri,
                        "skos:prefLabel": rec["ENTITY_DESC"],
                    })

                for rel in data["RELATED_ENTITIES"]:
                    match_key: str = rel["MATCH_KEY"]
                    match_level: int = rel["MATCH_LEVEL"]
                    match_code: str = rel["MATCH_LEVEL_CODE"]

                    why: str = f"{match_key} {match_level}"
                    pred_iri = "skos:related"

                    if match_code == "POSSIBLY_SAME":
                        pred_iri = "skos:closeMatch"

                    rel_list.append({
                        "pred": pred_iri,
                        "obj": self.SZ_PREFIX + str(rel["ENTITY_ID"]),
                        "skos:definition": why,
                    })

                ent_node: dict = {
                    "iri": entity_id,
                    "skos:prefLabel": ent_descrip,
                }

                if debug:
                    ic(ent_node)

                rdf_frag: str = f"{entity_id} skos:prefLabel \"{ent_descrip}\"@{language} "

                lemma_key: str = self.parse_lemma(simple_pipe(ent_descrip))  # type: ignore
                rdf_frag += f";\n  strw:lemma_phrase \"{lemma_key}\"@{language} "

                for rec_node in rec_list:
                    dat_rec: dict = data_records[rec_node["obj"]]
                    ent_type = dat_rec["RECORD_TYPE"]
                    rdf_frag += f';\n  {rec_node["pred"]} {rec_node["obj"]} '

                    if ent_type == "ORGANIZATION":
                        org_map[rec_node["skos:prefLabel"]] = entity_id

                for rel_node in rel_list:
                    rdf_frag += f';\n  {rel_node["pred"]} {rel_node["obj"]} '

                rdf_frag += f";\n  rdf:type strw:SzEntity, strw:{ent_type.capitalize()} "
                rdf_frag += "\n."
                rdf_list.append(rdf_frag)

        # construct the RDF graph
        for record_id, rec in data_records.items():
            rec_type: str = f'strw:{rec["RECORD_TYPE"].capitalize()}'
            name: str = ""
            employer: str = ""
            urls: typing.List[ str ] = []

            if rec_type == "strw:Organization":
                name = rec["NAMES"][0]["PRIMARY_NAME_ORG"]

                if "LINKS" in rec:
                    for url_dict in rec["LINKS"]:
                        for url in url_dict.values():
                            urls.append(url)

                if "WEBSITE_ADDRESS" in rec:
                    urls.append(rec["WEBSITE_ADDRESS"])

            else:
                if "NAME_FIRST" in rec:
                    name = rec["NAME_FIRST"]

                if "NAME_MIDDLE" in rec:
                    name += " " + rec["NAME_MIDDLE"]

                if "NAME_LAST" in rec:
                    name += " " + rec["NAME_LAST"]

                if "SOURCE_LINKS" in rec:
                    for url_dict in rec["SOURCE_LINKS"]:
                        for url in url_dict.values():
                            urls.append(url)

                if "EMPLOYER_NAME" in rec:
                    org_name: str = rec["EMPLOYER_NAME"]

                    if org_name in org_map:
                        employer = org_map.get(org_name)  # type: ignore

            rdf_frag = f"{record_id} rdf:type strw:DataRecord, {rec_type} "
            rdf_frag += f";\n  skos:prefLabel \"{name}\"@{language} "

            lemma_key = self.parse_lemma(simple_pipe(name))  # type: ignore
            rdf_frag += f";\n  strw:lemma_phrase \"{lemma_key}\"@{language} "

            for url in urls:
                rdf_frag += f";\n  dct:identifier <{url}> "

            rdf_frag += "\n."
            rdf_list.append(rdf_frag)

            if len(employer) > 0:
                rdf_frag = f"{parent[record_id]} org:memberOf {employer} ."
                rdf_list.append(rdf_frag)

        # serialize the generated RDF file
        with open(rdf_path, "w", encoding = "utf-8") as fp:
            fp.write("\n".join(rdf_list))

        # load the RDF graph
        rdf_graph: rdflib.Graph = rdflib.Graph()
        rdf_graph.parse(
            rdf_path.as_posix(),
            format = "turtle",
        )


    def populate_er_node (
        self,
        er_graph: rdflib.Graph,
        entity_iri: rdflib.term.Node,
        ) -> int:
        """
Populate a semantic layer node from an ER entity.
        """
        lemma_phrase_iri: rdflib.term.URIRef = self.rel_iri(StrwVocab.LEMMA_PHRASE)

        lemmas: typing.List[ str ] = [
            lemma.toPython()  # type: ignore
            for lemma in er_graph.objects(entity_iri, lemma_phrase_iri)
        ]

        lemma_key: str = lemmas[0]
        self.add_lemma(lemma_key)
        node_id: int = self.get_lemma_index(lemma_key)
        label: str = entity_iri.n3(er_graph.namespace_manager)

        text: str = er_graph.value(
            entity_iri,
            SKOS.prefLabel,
        ).toPython()  # type: ignore

        self.sem_layer.add_node(
            node_id,
            kind = NodeKind.ENTITY.value,
            key = lemma_key,
            label = label,
            text = text,
            iri = entity_iri,
            rank = 0.0,
            count = 1,
        )

        return node_id


    def load_er_thesaurus (  # pylint: disable=R0914
        self,
        datasets: typing.List[ str ],
        er_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Iterate through the _entity resolution_ results, adding a
domain-specific thesaurus of entities and relations into the
semantic layer.
        """
        self.parse_er_export(datasets)

        node_map: typing.Dict[ str, int ] = {}

        # load the ER triples into their own graph, to extrant and
        # link the known lemmas (i.e., the synonyms in the thesaurus)
        if er_path is None:
            er_path = pathlib.Path(self.config["er"]["thesaurus_path"])

        er_graph: rdflib.Graph = rdflib.Graph()

        er_graph.parse(
            er_path.as_posix(),
            format = "turtle",
        )

        # first iterate through the data records, loading lemma keys
        # and populating nodes in the semantic layer
        for entity_iri in er_graph.subjects(RDF.type, self.form_concept("DataRecord")):
            node_id = self.populate_er_node(er_graph, entity_iri)
            node_map[entity_iri.n3(er_graph.namespace_manager)] = node_id

        # now iterate through the entities, overriding any prior lemma
        # keys from data records
        for entity_iri in er_graph.subjects(RDF.type, self.form_concept("SzEntity")):
            node_id = self.populate_er_node(er_graph, entity_iri)
            node_map[entity_iri.n3(er_graph.namespace_manager)] = node_id

        # then add SKOS relations (thesaurus synonyms and taxonymy)
        # as edges in the semantic layer
        for entity_iri in er_graph.subjects(RDF.type, self.form_concept("SzEntity")):
            for sem_rel in [ SKOS.related, SKOS.closeMatch, SKOS.exactMatch, ORG.memberOf ]:
                for obj in er_graph.objects(entity_iri, sem_rel):
                    src_id: int = node_map[entity_iri.n3(er_graph.namespace_manager)]
                    dst_id: int = node_map[obj.n3(er_graph.namespace_manager)]

                    if src_id != dst_id:
                        rel_iri: str = sem_rel.n3(er_graph.namespace_manager)
                        prob: float = 0.5

                        if rel_iri in [ "skos:exactMatch", "org:memberOf" ]:
                            prob = 1.0

                        self.sem_layer.add_edge(
                            src_id,
                            dst_id,
                            key = rel_iri,
                            prob = prob,
                        )

        # also link entities to their taxonomy nodes
        for taxo_iri in [ self.form_concept("Organization"), self.form_concept("Person") ]:
            for entity_iri in er_graph.subjects(RDF.type, taxo_iri):
                node_id = node_map[entity_iri.n3(er_graph.namespace_manager)]
                taxo_node_id: int = self.taxo_node[taxo_iri.n3(self.rdf_graph.namespace_manager)]

                self.sem_layer.add_edge(
                    node_id,
                    taxo_node_id,
                    key = "RDF:type",
                    weight = 0.0
                )

        # finally, load ER triples into the semantic layer
        self.rdf_graph.parse(
            er_path.as_posix(),
            format = "turtle",
        )


    def get_ner_labels (
        self,
        ) -> typing.List[ str ]:
        """
Iterate through `skos:Concept` entities to extract the labels used for
zero-shot NER.
        """
        return [
            label.toPython()  # type: ignore
            for concept_iri in self.rdf_graph.subjects(RDF.type, SKOS.Concept)
            for label in self.rdf_graph.objects(concept_iri, SKOS.prefLabel, unique = True)
        ]


    def get_label_map (
        self,
        ) -> typing.Dict[ str, str ]:
        """
Iterate through `skos:Concept` entities to extract a mapping between
NER labels and abbreviated IRIs.
        """
        return {
            label.toPython(): concept_iri.n3(self.rdf_graph.namespace_manager)  # type: ignore
            for concept_iri in self.rdf_graph.subjects(RDF.type, SKOS.Concept)
            for label in self.rdf_graph.objects(concept_iri, SKOS.prefLabel, unique = True)
        }


    def vis_nodes (
        self,
        num_docs: int,
        ) -> typing.Iterator[ typing.Tuple[ int, dict ]]:
        """
Iterator for the visualization attributes of nodes.
        """
        for node_id, node_attr in self.sem_layer.nodes(data = True):
            attr: dict = {}

            if node_attr.get("kind") == NodeKind.ENTITY.value and node_attr.get("label") not in [ "NP" ]:  # pylint: disable=C0301
                attr["color"] = "hsla(65, 46%, 58%, 0.80)"
                attr["size"] = round(20 * math.log(1.0 + math.sqrt(float(node_attr.get("count"))) / num_docs))  # type: ignore # pylint: disable=C0301
                attr["label"] = node_attr.get("text")  # type: ignore
                attr["title"] = node_attr.get("key")  # type: ignore
            elif node_attr.get("kind") == NodeKind.TAXONOMY.value:
                attr["color"] = "hsla(306, 45%, 57%, 0.95)"
                attr["size"] = 5
                attr["label"] = node_attr.get("label")  # type: ignore
                attr["title"] = node_attr.get("iri")  # type: ignore
            else:
                continue

            yield node_id, attr


    def vis_edges (
        self,
        ) -> typing.Iterator[ typing.Tuple[ int, int, str ]]:
        """
Iterator for the visualization attributes of edges.
        """
        for src_node, dst_node, key in self.sem_layer.edges(keys = True):
            yield src_node, dst_node, key


    def add_entity_sequence (
        self,
        span_decoder: typing.Dict[ tuple, Entity ],
        ) -> None:
        """
Build the vector input for entity embeddings.
        """
        w2v_map: typing.Dict[ int, typing.Set[ str ]] = defaultdict(set)

        for ent in span_decoder.values():
            if ent.node is not None:
                w2v_map[ent.sent_id].add(ent.key)

        for sent_id, ents in w2v_map.items():
            vec: list = list(ents)
            vec.insert(0, str(sent_id))
            self.w2v_vectors.append(vec)


    def embed_entities (
        self,
        *,
        w2v_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Train a `gensim.Word2Vec` model for entity embeddings.
        """
        w2v_max: int = max([  # pylint: disable=R1728
            len(vec) - 1
            for vec in self.w2v_vectors
        ])

        self.w2v_model = gensim.models.Word2Vec(
            self.w2v_vectors,
            min_count = 2,
            window = w2v_max,
        )

        if w2v_path is None:
            w2v_path = pathlib.Path(self.config["ent"]["w2v_path"])

        self.w2v_model.save(w2v_path.as_posix())


    def save_sem_layer (
        self,
        *,
        kg_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Serialize the constructed KG as a JSON file represented in the
_node-link_ data format.

Aternatively this could be stored in a graph database.
        """
        if kg_path is None:
            kg_path = pathlib.Path(self.config["kg"]["kg_path"])

        with kg_path.open("w", encoding = "utf-8") as fp:
            fp.write(
                json.dumps(
                    nx.node_link_data(
                        self.sem_layer,
                        edges = "edges",
                    ),
                    indent = 2,
                    sort_keys = True,
                )
            )


    def save_assets (
        self,
        *,
        kg_path: typing.Optional[ pathlib.Path ] = None,
        w2v_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Serialize the assets for reusing a constructed KG.
        """
        self.embed_entities(w2v_path = w2v_path)
        self.save_sem_layer(kg_path = kg_path)


    def load_assets (
        self,
        *,
        kg_path: typing.Optional[ pathlib.Path ] = None,
        w2v_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Load the serialized assets for a previoulys constructed KG.
        """
        vect_db: lancedb.db.LanceDBConnection = lancedb.connect(self.config["vect"]["lancedb_uri"])
        self.chunk_table = vect_db.open_table(self.config["vect"]["chunk_table"])

        if w2v_path is None:
            w2v_path = pathlib.Path(self.config["ent"]["w2v_path"])

        self.w2v_model = gensim.models.Word2Vec.load(w2v_path.as_posix())

        if kg_path is None:
            kg_path = pathlib.Path(self.config["kg"]["kg_path"])

        with pathlib.Path(kg_path).open("r", encoding = "utf-8") as fp:
            self.sem_layer = nx.node_link_graph(
                json.load(fp),
                edges = "edges",
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Construct the lexical graph and condense it into a knowledge graph.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import typing

from icecream import ic  # type: ignore
import networkx as nx
import polars as pl
import spacy

from .context import DomainContext
from .elem import Entity, NodeKind, StrwVocab, TextChunk
from .nlp import Parser
from .scrape import Scraper
from .textrank import run_textrank, cooccur_entities


class KnowledgeGraph:
    """
Construct a _knowledge graph_ and build out assets to serialize then
use later.
    """

    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config


    def build_graph (  # pylint: disable=R0912,R0913,R0914,R0917,W0102
        self,
        url_list: typing.List[ str ],
        domain_context: DomainContext,
        parser: Parser,
        simple_pipe: spacy.Language,
        entity_pipe: spacy.Language,
        *,
        debug: bool = False,
        ) -> None:
        """
Construct a knowledge graph from unstructured data sources.
        """
        label_map: typing.Dict[ str, str ] = domain_context.get_label_map()

        # iterate through the URL list, scraping text and building chunks
        scraper: Scraper = Scraper(self.config, parser)
        chunk_id: int = domain_context.start_chunk_id

        for url in url_list:
            # define data structures intialized for each parsed document
            lex_graph: nx.MultiDiGraph = nx.MultiDiGraph()
            chunk_list: typing.List[ TextChunk ] = []

            chunk_id = scraper.scrape_html(
                simple_pipe,
                url,
                chunk_list,
                chunk_id,
            )

            domain_context.chunk_table.add(chunk_list)  # type: ignore

            # parse each chunk to build a lexical graph per source URL
            for chunk in chunk_list:
                span_decoder: typing.Dict[ tuple, Entity ] = {}

                doc: spacy.tokens.doc.Doc = parser.parse_text(  # pylint: disable=I1101
                    domain_context,
                    entity_pipe,
                    lex_graph,
                    chunk,
                    debug = debug,
                )

                if debug:
                    ic(chunk)

                # keep track of sentence numbers per chunk, to use later
                # for entity co-occurrence links
                sent_map: typing.Dict[ spacy.tokens.span.Span, int ] = {}  # pylint: disable=I1101

                for sent_id, sent in enumerate(doc.sents):
                    sent_map[sent] = sent_id

                # classify the recognized spans within this chunk as
                # potential entities

                # NB: if we'd run [_entity resolution_]
                # see: <https://neo4j.com/developer-blog/entity-resolved-knowledge-graphs/>
                # previously from _structured_ or _semi-structured_ data sources to
                # generate a "backbone" for the knowledge graph, then we could use
                # contextualized _surface forms_ perform _entity linking_ on the
                # entities extracted here from _unstructured_ data

                for span in doc.ents:
                    self.make_entity(
                        domain_context,
                        span_decoder,
                        sent_map,
                        span,
                        label_map[span.label_], # decoded as abbrev IRI
                        chunk,
                        debug = debug,
                    )

                for span in doc.noun_chunks:
                    self.make_entity(
                        domain_context,
                        span_decoder,
                        sent_map,
                        span,
                        "NP",
                        chunk,
                        debug = False, # debug
                    )

                # overlay the recognized entity spans atop the base layer
                # constructed by _textgraph_ analysis of the `spaCy` parse trees
                for ent in span_decoder.values():
                    if ent.key not in parser.STOP_WORDS:
                        self.extract_entity(
                            domain_context,
                            lex_graph,
                            ent,
                            debug = debug,
                        )

                # extract relations for co-occurring entity pairs
                ## PLACEHOLDER

                # connect entities which co-occur within the same sentence
                cooccur_entities(
                    lex_graph,
                    span_decoder,
                )

                domain_context.add_entity_sequence(span_decoder)

            # apply _textrank_ to the graph (in the url/doc iteration)
            # then report the top-ranked extracted entities
            df: pl.DataFrame = run_textrank(
                self.config,
                lex_graph,
            )

            if debug:
                ic(url, df.head(11))

            # abstract a semantic overlay from the lexical graph
            # and persist this in the resulting KG
            self.abstract_overlay(
                domain_context,
                url,
                chunk_list,
                lex_graph,
            )

            if debug:
                print(
                    "nodes",
                    len(domain_context.sem_layer.nodes),
                    "edges",
                    len(domain_context.sem_layer.edges),
                )


    def abstract_overlay (  # pylint: disable=R0912,R0914
        self,
        domain_context: DomainContext,
        url: str,
        chunk_list: typing.List[ TextChunk ],
        lex_graph: nx.MultiDiGraph,
        ) -> None:
        """
Abstract a _semantic overlay_ from the lexical graph -- in other words
which nodes and edges get promoted up to the next level?

Also connect the extracted entities with their source chunks, where
the latter first-class citizens within the KG.
        """
        kept_nodes: typing.Set[ int ] = set()

        skipped_rel: typing.Set[ str ] = set([
            StrwVocab.CO_OCCURS_WITH.value,
            StrwVocab.COMPOUND_ELEM_OF.value,
            StrwVocab.FOLLOWS_LEXICALLY.value,
        ])

        chunk_nodes: typing.Dict[ int, str ] = {
            chunk.uid: f"chunk_{chunk.uid}"
            for chunk in chunk_list
        }

        for chunk_id, node_id in chunk_nodes.items():
            domain_context.sem_layer.add_node(
                node_id,
                kind = NodeKind.CHUNK.value,
                chunk = chunk_id,
                url = url,
            )

        for node_id, node_attr in lex_graph.nodes(data = True):
            if node_attr["kind"] == NodeKind.ENTITY.value:
                kept_nodes.add(node_id)
                count: int = node_attr["count"]

                if not domain_context.sem_layer.has_node(node_id):
                    domain_context.sem_layer.add_node(
                        node_id,
                        kind = NodeKind.ENTITY.value,
                        key = node_attr["key"],
                        text = node_attr["text"],
                        label = node_attr["label"],
                        rank = round(node_attr["rank"], 4),
                        count = count,
                    )
                else:
                    domain_context.sem_layer.nodes[node_id]["count"] += count

                domain_context.sem_layer.add_edge(
                    node_id,
                    chunk_nodes[node_attr["chunk"]],
                    key = StrwVocab.WITHIN_CHUNK.value,
                    weight = round(node_attr["rank"], 4),
                )

                # link each entity to its taxonomy concept,
                # though be careful not to introduce cycles
                if node_attr["label"] not in [ "NP" ]:
                    taxo_node_id: int = domain_context.taxo_node[node_attr["label"]]

                    if node_id != taxo_node_id:
                        domain_context.sem_layer.add_edge(
                            node_id,
                            taxo_node_id,
                            key = "RDF:type",
                            weight = 0.0
                        )

        for src_id, dst_id, key, edge_attr in lex_graph.edges(data = True, keys = True):
            if src_id in kept_nodes and dst_id in kept_nodes:
                prob: float = 1.0

                if "prob" in edge_attr:
                    prob = edge_attr["prob"]

                if key not in skipped_rel:
                    if not domain_context.sem_layer.has_edge(src_id, dst_id):
                        domain_context.sem_layer.add_edge(
                            src_id,
                            dst_id,
                            key = key,
                            prob = prob,
                        )
                    else:
                        domain_context.sem_layer.edges[src_id, dst_id, key]["prob"] = max(
                            prob,
                            domain_context.sem_layer.edges[src_id, dst_id, key]["prob"],
                        )


    def make_entity (  # pylint: disable=R0913,R0917
        self,
        domain_context: DomainContext,
        span_decoder: typing.Dict[ tuple, Entity ],
        sent_map: typing.Dict[ spacy.tokens.span.Span, int ],  # pylint: disable=I1101
        span: spacy.tokens.span.Span,  # pylint: disable=I1101
        label: str,
        chunk: TextChunk,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> Entity:
        """
Instantiate one `Entity` object, adding to our working "vocabulary".
        """
        lemma_key: str = domain_context.parse_lemma(span)  # type: ignore

        ent: Entity = Entity(
            ( span.start, span.end, ),
            lemma_key,
            span.text,
            label,
            chunk.uid,
            sent_map[span.sent],
            span,
        )

        if ent.loc not in span_decoder:
            span_decoder[ent.loc] = ent

            if False: # debug  # pylint: disable=W0125
                ic(ent)

        return ent


    def extract_entity (
        self,
        domain_context: DomainContext,
        lex_graph: nx.MultiDiGraph,
        ent: Entity,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> None:
        """
Link one `Entity` into this doc's lexical graph.
        """
        prev_known: bool = domain_context.add_lemma(ent.key)
        node_id: int = domain_context.get_lemma_index(ent.key)
        ent.node = node_id

        # hydrate a compound phrase in this doc's lexical graph
        if not lex_graph.has_node(node_id):
            lex_graph.add_node(
                node_id,
                key = ent.key,
                kind = NodeKind.ENTITY.value,
                label = ent.label,
                pos = "NP",
                text = ent.text,
                chunk = ent.chunk_id,
                count = 1,
            )

            for tok in ent.span:
                tok_lemma_key: str = domain_context.parse_lemma([ tok ])  # type: ignore

                if tok_lemma_key in domain_context.known_lemma:
                    tok_idx: int = domain_context.get_lemma_index(tok_lemma_key)

                    lex_graph.add_edge(
                        node_id,
                        tok_idx,
                        key = StrwVocab.COMPOUND_ELEM_OF.value,
                    )

        if prev_known:
            # promote a previous Lemma node to an Entity
            node: dict = lex_graph.nodes[node_id]
            node["kind"] = NodeKind.ENTITY.value
            node["chunk"] = ent.chunk_id
            node["count"] += 1

            # select the more specific label
            if "label" not in node or node["label"] == "NP":
                node["label"] = ent.label

        if False: # debug  # pylint: disable=W0125
            ic(ent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Interactive visualization of a `NetworkX` graph based on `PyVis` to
generate an HTML file. Subclass this interface to integrate other
visualization libraries.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import typing

import pyvis  # type: ignore


class VisHTML:  # pylint: disable=R0902
    """
Implementation for `PyVis` to generate an HTML visualization.
    """

    def __init__ (
        self,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = {}


    def set_config (
        self,
        config: dict,
        ) -> None:
        """
Accessor method to configure -- part of a design pattern to make the
iteractive visualization more "pluggable", i.e., to be subclassed and
customized for other visualization libraries.
        """
        self.config = config


    def gen_vis_html (  # pylint: disable=R0914
        self,
        html_file: str,
        node_iter: typing.Iterator[ typing.Tuple[ int, dict ]],
        edge_iter: typing.Iterator[ typing.Tuple[ int, int, str ]],
        *,
        notebook: bool = False,
        ) -> None:
        """
Use `pyvis` to provide an interactive visualization of the graph layers.
        """
        kept_nodes: typing.Set[ int ] = set()

        pv_net: pyvis.network.Network = pyvis.network.Network(
            height = self.config["vis"]["html_height"],
            width = self.config["vis"]["html_width"],
            notebook = notebook,
            cdn_resources = "remote",
        )

        for node_id, attr in node_iter:
            kept_nodes.add(node_id)

            pv_net.add_node(
                node_id,
                label = attr["label"],
                title = attr["title"],
                color = attr["color"],
                size = attr["size"],
            )

        for src_node, dst_node, key in edge_iter:
            if src_node in kept_nodes and dst_node in kept_nodes:
                pv_net.add_edge(
                    src_node,
                    dst_node,
                    title = key,
                )

        pv_net.toggle_physics(True)
        pv_net.show_buttons(filter_ = [ "physics" ])
        pv_net.save_graph(html_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Builds assets for constructing a KG, then running GraphRAG downstream.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import itertools
import logging
import os
import pathlib
import tomllib
import traceback
import typing
import warnings

from icecream import ic  # type: ignore
import networkx as nx
import polars as pl
import spacy
import transformers  # type: ignore

from .baml_client import b
from .baml_client import types as baml_types
from .context import DomainContext
from .elem import NodeKind
from .kg import KnowledgeGraph
from .nlp import Parser
from .vis import VisHTML


class Strwythura:  # pylint: disable=R0902
    """
Builds assets for constructing a KG, then running GraphRAG downstream.
    """

    def __init__ (
        self,
        domain_context: DomainContext,
        *,
        config_path: pathlib.Path = pathlib.Path("config.toml"),
        overwrite: bool = False,
        ) -> None:
        """
Constructor.
        """
        # configuration
        self.config: dict = {}

        with open(config_path, mode = "rb") as fp:
            self.config = tomllib.load(fp)

        # disable noisy logging
        os.environ["BAML_LOG"] = "WARN"
        os.environ["TOKENIZERS_PARALLELISM"] = "0"

        logging.disable(logging.ERROR)
        transformers.logging.set_verbosity_error()

        ## none of this works!
        #os.environ["TQDM_DISABLE"] = "1"
        #loguru.logger.disable(gliner_spacy.pipeline.__name__)
        #loggers: dict =
        # { name:logging.getLogger(name) for name in logging.root.manager.loggerDict }
        #ic(loggers)

        # initialize the data structures used for assets
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            self.domain_context: DomainContext = domain_context
            self.domain_context.set_config(self.config)

            if overwrite:
                self.domain_context.init_chunk_table()
                self.domain_context.load_taxonomy()

            self.parser: Parser = Parser(self.config)
            self.simple_pipe: spacy.Language = spacy.load(self.config["nlp"]["spacy_model"])
            self.entity_pipe: spacy.Language = self.parser.build_entity_pipe(
                self.domain_context.get_ner_labels()
            )


    def build_assets (
        self,
        url_list: typing.List[ str ],
        *,
        debug: bool = False,
        ) -> None:
        """
Builds assets for constructing a KG.
        """
        try:
            kg: KnowledgeGraph = KnowledgeGraph(self.config)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                kg.build_graph(
                    url_list,
                    self.domain_context,
                    self.parser,
                    self.simple_pipe,
                    self.entity_pipe,
                    debug = debug,
                )

        except Exception as ex:  # pylint: disable=W0718
            ic(ex)
            traceback.print_exc()


    def gen_visualization (
        self,
        url_list: typing.List[ str ],
        vis_html: VisHTML,
        *,
        html_path: typing.Optional[ pathlib.Path ] = None,
        ) -> None:
        """
Generate HTML for an interactive visualization of the graph.
        """
        if html_path is None:
            html_path = pathlib.Path(self.config["kg"]["html_path"])

        vis_html.set_config(self.config)

        vis_html.gen_vis_html(
            html_path.as_posix(),
            self.domain_context.vis_nodes(len(url_list)),
            self.domain_context.vis_edges(),
        )


class GraphRAG:
    """
Run an example query through LanceDB to identify _chunks_ and through
the Word2Vec entity embedding model for a _semantic expansion_ to
produce a set of _anchor nodes_ in the NetworkX graph.
    """

    def __init__ (
        self,
        strw: Strwythura,
        ) -> None:
        """
Constructor.
        """
        self.strw: Strwythura = strw


    def find_entities (
        self,
        question: str,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ str ]:
        """
Extract entity spans from a text question, generating their lemma keys.
        """
        doc: spacy.tokens.doc.Doc = self.strw.entity_pipe(question)  # type: ignore  # pylint: disable=I1101

        for span in doc.ents:
            lemma_key: str = self.strw.domain_context.parse_lemma(span)  # type: ignore

            if debug:
                ic(span, lemma_key)

            yield lemma_key


    def get_chunks (  # pylint: disable=R0912,R0914
        self,
        question: str,
        *,
        debug: bool = False,
        num_chunks: typing.Optional[ int ] = None,
        ) -> typing.List[ str ]:
        """
Run semantic search to produce a set of text chunks.
        """
        chunk_ids: typing.List[ int ] = []

        if num_chunks is None:
            num_chunks = self.strw.config["rag"]["num_chunks"]

        # extract entities from the question
        entities: typing.Set[ str ] = set(list(self.find_entities(question)))

        if debug:
            ic(entities)

        # semantic expansion using entity embeddings
        neighbors: typing.Set[ str ] = set()

        try:
            for entity in entities:
                neighbor_iter = self.strw.domain_context.w2v_model.wv.most_similar(  # type: ignore
                    positive = [ entity ],
                    topn = num_chunks,
                )

                for neighbor in neighbor_iter:
                    neighbors.add(neighbor)
        except KeyError:
            pass

        if debug:
            ic(neighbors)

        # map the expanded set of entities to nodes in the graph
        expanded_entities: set = entities.union(neighbors)

        anchor_nodes: set = {
            node
            for node, dat in self.strw.domain_context.sem_layer.nodes(data = True)
            if "key" in dat and dat["key"] in expanded_entities
        }

        if debug:
            ic(anchor_nodes)

        # trap edge cases where there is no connected subgraph
        try:
            # extract a subgraph based on shortest paths between anchor nodes
            node_iter: typing.Iterator[ int ] = self.extract_subgraph(
                anchor_nodes,
                debug = debug,
            )

            # extract the chunk neighbors
            chunk_iter: typing.Iterator[ int ] = self.extract_chunk_neighbors(
                list(node_iter),
                debug = debug,
            )

            for chunk_id in chunk_iter:
                if chunk_id not in chunk_ids:
                    chunk_ids.append(chunk_id)
        except nx.exception.NetworkXNoPath:
            pass

        # enumerate chunks from a vector search -- the basic RAG process
        df_ann_chunk: pl.DataFrame = self.strw.domain_context.chunk_table.search(  # type: ignore
            question
        ).limit(
            num_chunks
        ).to_polars()

        for row in df_ann_chunk.iter_rows(named = True):
            chunk_id: int =  row["uid"]  # type: ignore

            if chunk_id not in chunk_ids:
                chunk_ids.append(chunk_id)

        if debug:
            ic(chunk_ids)

        # get the text for the combined/ranked list of chunks
        id_list: str = ", ".join([ str(c_id) for c_id in chunk_ids ])
        filter_term: str = f"uid IN ({id_list})"

        chunks: typing.List[ str ] = self.strw.domain_context.chunk_table.search().where(  # type: ignore  # pylint: disable=C0301
            filter_term
        ).select(
            [ "text" ]
        ).to_polars()["text"].to_list()

        return chunks[:num_chunks]


    def gen_subgraph_paths (
        self,
        anchor_nodes: set,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ str ]:
        """
Generate pairwise shortest paths among the nodes from semantic
expansion, to define a subgraph.

In other words, this emulates a _semantic random walk_.
        """
        for pair in itertools.combinations(anchor_nodes, 2):
            if debug:
                ic(pair)

            for path in nx.all_shortest_paths(self.strw.domain_context.sem_layer, pair[0], pair[1]):
                if debug:
                    ic(path)

                for node in path:
                    if node not in pair:
                        dat: dict = self.strw.domain_context.sem_layer.nodes[node]

                        if debug:
                            ic(node, dat)

                        yield node


    def extract_subgraph (
        self,
        anchor_nodes: set,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ int ]:
        """
Extract a subgraph, run a _centrality_ algorithm to rerank the most
referenced entities in the subgraph.
        """
        subgraph_iter: typing.Iterator[ str ] = self.gen_subgraph_paths(
            anchor_nodes,
            debug = debug,
        )

        subgraph: nx.MultiDiGraph = self.strw.domain_context.sem_layer.subgraph(  # type: ignore
            anchor_nodes.union(set(subgraph_iter))
        )

        rank_iter: dict = nx.pagerank(  # type: ignore
            subgraph,
            self.strw.config["tr"]["tr_alpha"],
        ).items()

        for node, rank in sorted(rank_iter, key = lambda x: x[1], reverse = True):
            dat: dict = self.strw.domain_context.sem_layer.nodes[node]

            if debug:
                ic(node, rank, dat)

            yield node


    def extract_chunk_neighbors (
        self,
        ranked_nodes: list,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ int ]:
        """
Find the neighboring chunks for each _anchor node_ in the given list.
        """
        for node in ranked_nodes:
            if debug:
                ic(node)

            for neighbor in self.strw.domain_context.sem_layer.neighbors(node):
                dat: dict = self.strw.domain_context.sem_layer.nodes[neighbor]

                if dat["kind"] == NodeKind.CHUNK.value:
                    if debug:
                        ic(neighbor, dat)

                    chunk_id: int = int(neighbor.replace("chunk_", ""))
                    yield chunk_id


    def qa_cycle (
        self,
        question: str,
        *,
        debug: bool = False,
        ) -> baml_types.Response:
        """
Loop to answer questions.
        """
        chunks: typing.List[ str ] = self.get_chunks(
            question,
            debug = debug,
        )

        context: str = "\n".join( chunks )

        response: baml_types.Response = b.RAG(
            question,
            context,
        )

        return response



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
from baml_py import type_builder
from baml_py import baml_py
# These are exports, not used here, hence the linter is disabled
from baml_py.baml_py import FieldType, EnumValueBuilder, EnumBuilder, ClassBuilder # noqa: F401 # pylint: disable=unused-import
from .globals import DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_RUNTIME

class TypeBuilder(type_builder.TypeBuilder):
    def __init__(self):
        super().__init__(classes=set(
          ["Response",]
        ), enums=set(
          []
        ), runtime=DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_RUNTIME)

    # #########################################################################
    # Generated enums 0
    # #########################################################################


    # #########################################################################
    # Generated classes 1
    # #########################################################################

    @property
    def Response(self) -> "ResponseViewer":
        return ResponseViewer(self)



# #########################################################################
# Generated enums 0
# #########################################################################


# #########################################################################
# Generated classes 1
# #########################################################################

class ResponseAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("Response")
        self._properties: typing.Set[str] = set([  "question_type",  "question",  "answer",  ])
        self._props = ResponseProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "ResponseProperties":
        return self._props


class ResponseViewer(ResponseAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def list_properties(self) -> typing.List[typing.Tuple[str, type_builder.ClassPropertyViewer]]:
        return [(name, type_builder.ClassPropertyViewer(self._bldr.property(name))) for name in self._properties]
    


class ResponseProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    
    @property
    def question_type(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("question_type"))
    
    @property
    def question(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("question"))
    
    @property
    def answer(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("answer"))
    
    




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
NLP methods for constructing the _lexical graph_.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import typing
import unicodedata

from gliner_spacy.pipeline import GlinerSpacy  # type: ignore # pylint: disable=W0611
from icecream import ic  # type: ignore
import networkx as nx
import spacy
import w3lib.html

from .context import DomainContext
from .elem import StrwVocab, TextChunk


class Parser:
    """
Wrapper class for the `spaCy` NLP pipeline used to extract entities
and relations, based on using `GLiNER`, BAML, and textgraphs.
    """
    STOP_WORDS: typing.Set[ str ] = set([
        "PRON.each",
        "PRON.he",
        "PRON.it",
        "PRON.she",
        "PRON.some",
        "PRON.someone",
        "PRON.that",
        "PRON.their",
        "PRON.they",
        "PRON.those",
        "PRON.we",
        "PRON.what",
        "PRON.which",
        "PRON.who",
        "PRON.you",
    ])


    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.ner_labels: typing.List[ str ] = []


    def build_entity_pipe (
        self,
        ner_labels: typing.List[ str ],
        ) -> spacy.Language:
        """
Initialize the `spaCy` pipeline used for NER + RE, by loading models
for `spaCy`, `GLiNER`

  - `ner_labels`: semantics to apply for zero-shot NER

Note: this may take several minutes when run the first time after
installing the repo.
        """
        self.ner_labels = ner_labels
        entity_pipe: spacy.Language = spacy.load(self.config["nlp"]["spacy_model"])

        entity_pipe.add_pipe(
            "gliner_spacy",
            config = {
                "style": "ent",
                "labels": self.ner_labels,
                "gliner_model": self.config["nlp"]["gliner_model"],
                "chunk_size": self.config["vect"]["chunk_size"],
            },
        )

        return entity_pipe


    def parse_text (  # pylint: disable=R0913,R0914
        self,
        domain_context: DomainContext,
        entity_pipe: spacy.Language,
        lex_graph: nx.MultiDiGraph,
        chunk: TextChunk,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> spacy.tokens.doc.Doc:  # pylint: disable=I1101
        """
Parse an input text chunk, returning a `spaCy` document.
        """
        doc: spacy.tokens.doc.Doc = entity_pipe(chunk.text)  # type: ignore  # pylint: disable=I1101

        # scan the document tokens to add lemmas into the _lexical graph_
        # then use a _textgraph_ distillation approach called "textrank"
        for sent in doc.sents:
            node_seq: typing.List[ int ] = []

            if False: # debug  # pylint: disable=W0125
                ic(sent)

            for tok in sent:
                text: str = tok.text.strip()
                pos: str = tok.pos_

                if pos in DomainContext.POS_TRANSFORM:
                    pos = DomainContext.POS_TRANSFORM[pos]

                if pos == "NOUN":
                    lemma_key: str = domain_context.parse_lemma([ tok ])  # type: ignore
                    prev_known: bool = domain_context.add_lemma(lemma_key)
                    node_id: int = domain_context.get_lemma_index(lemma_key)
                    node_seq.append(node_id)

                    if not lex_graph.has_node(node_id):
                        lex_graph.add_node(
                            node_id,
                            key = lemma_key,
                            kind = "Lemma",
                            pos = pos,
                            text = text,
                            chunk = chunk,
                            count = 1,
                        )

                    elif prev_known:
                        node: dict = lex_graph.nodes[node_id]
                        node["count"] += 1

            # create the _textrank_ edges for the lexical graph,
            # which will get used for ranking, but discarded later
            if False: # debug  # pylint: disable=W0125
                ic(node_seq)

            for hop in range(self.config["tr"]["tr_lookback"]):
                for node_id, node in enumerate(node_seq[: -1 - hop]):  # type: ignore
                    neighbor: int = node_seq[hop + node_id + 1]

                    if not lex_graph.has_edge(node, neighbor):
                        lex_graph.add_edge(
                            node,
                            neighbor,
                            key = StrwVocab.FOLLOWS_LEXICALLY.value,
                        )

        return doc


    def uni_scrubber (
        self,
        span: spacy.tokens.span.Span,  # pylint: disable=I1101
        ) -> str:
        """
Applies multiple approaches for aggressively removing garbled Unicode
and spurious punctuation from the given text.

OH: "It scrubs the garble from its stream... or it gets the debugger again!"
        """
        text: str = span.text

        if type(text).__name__ != "str":
            print("not a string?", type(text), text)

        limpio: str = " ".join(map(lambda s: s.strip(), text.split("\n"))).strip()
        limpio = w3lib.html.replace_escape_chars(limpio)

        limpio = limpio.replace('“', '"').replace('”', '"')
        limpio = limpio.replace("‘", "'").replace("’", "'").replace("`", "'").replace("â", "'")
        limpio = limpio.replace("…", "...").replace("–", "-")

        limpio = str(unicodedata.normalize("NFKD", limpio).encode("ascii", "ignore").decode("utf-8"))  # pylint: disable=C0301

        return limpio


    def make_chunk (
        self,
        doc: spacy.tokens.doc.Doc,  # pylint: disable=I1101
        url: str,
        chunk_list: typing.List[ TextChunk ],
        chunk_id: int,
        ) -> int:
        """
Split the given document into text chunks, returning the last index.
BTW, for ideal text chunk size see
<https://www.llamaindex.ai/blog/evaluating-the-ideal-chunk-size-for-a-rag-system-using-llamaindex-6207e5d3fec5>
        """
        chunks: typing.List[ str ] = []
        chunk_total: int = 0
        prev_line: str = ""
        sent_id: int = 0

        for sent_id, sent in enumerate(doc.sents):
            line: str = self.uni_scrubber(sent)
            line_len: int = len(line)

            if (chunk_total + line_len) > self.config["vect"]["chunk_size"]:
                # emit the current chunk
                chunk_list.append(
                    TextChunk(
                        uid = chunk_id,
                        url = url,
                        sent_id = sent_id,
                        text = "\n".join(chunks),
                    )
                )

                # start a new chunk
                chunks = [ prev_line, line ]
                chunk_total = len(prev_line) + line_len
                chunk_id += 1
            else:
                # append line to the current chunk
                chunks.append(line)
                chunk_total += line_len

            prev_line = line

        # emit the trailing chunk
        chunk_list.append(
            TextChunk(
                uid = chunk_id,
                url = url,
                sent_id = sent_id + 1,
                text = "\n".join(chunks),
            )
        )

        return chunk_id + 1



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

from . import types
from . import stream_types


type_map = {

    "types.Response": types.Response,
    "stream_types.Response": stream_types.Response,


}


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Numerical utilities used for optimization.
see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import math
import typing

import numpy as np
import pandas as pd


def calc_quantile_bins (
    num_rows: int,
    *,
    amplitude: int = 4,
    ) -> np.ndarray:
    """
Calculate the bins to use for a quantile stripe,
using [`numpy.linspace`](https://numpy.org/doc/stable/reference/generated/numpy.linspace.html)

    num_rows:
number of rows in the target dataframe

    returns:
calculated bins, as a `numpy.ndarray`
    """
    granularity = max(round(math.log(num_rows) * amplitude), 1)

    return np.linspace(
        0,
        1,
        num = granularity,
        endpoint = True,
    )


def stripe_column (
    values: list,
    bins: int,
    ) -> np.ndarray:
    """
Stripe a column in a dataframe, by interpolating quantiles into a set of discrete indexes.

    values:
list of values to stripe

    bins:
quantile bins; see [`calc_quantile_bins()`](#calc_quantile_bins-function)

    returns:
the striped column values, as a `numpy.ndarray`
    """
    s = pd.Series(values)
    q = s.quantile(bins, interpolation = "nearest")

    try:
        stripe = np.digitize(values, q) - 1
        return stripe
    except ValueError as ex:
        # should never happen?
        print("ValueError:", str(ex), values, s, q, bins)
        raise


def root_mean_square (
    values: typing.List[ float ]
    ) -> float:
    """
Calculate the [*root mean square*](https://mathworld.wolfram.com/Root-Mean-Square.html)
of the values in the given list.

    values:
list of values to use in the RMS calculation

    returns:
RMS metric as a float
    """
    s: float = sum(map(lambda x: float(x) ** 2.0, values))
    n: float = float(len(values))

    return math.sqrt(s / n)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Data validation classes for constructing knowledge graphs.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from dataclasses import dataclass
from enum import StrEnum
import typing

from lancedb.embeddings import get_registry, transformers  # type: ignore
from lancedb.pydantic import LanceModel, Vector  # type: ignore
import spacy


class NodeKind (StrEnum):
    """
Values for the `kind` property in graph nodes.
    """
    CHUNK = "Chunk"
    ENTITY = "Entity"
    TAXONOMY = "Taxonomy"


class StrwVocab (StrEnum):
    """
Values for relations in the `strw:` RDF vocabulary.
    """
    LEMMA_PHRASE = "strw:lemma_phrase"
    FOLLOWS_LEXICALLY = "strw:follows_lexically"
    CO_OCCURS_WITH = "strw:co_occurs_with"
    COMPOUND_ELEM_OF = "strw:compound_elem_of"
    WITHIN_CHUNK = "strw:within_chunk"


# Note: `LanceDB` requires that the embedding model be hard-coded, so far
EMBED_MODEL: str = "BAAI/bge-small-en-v1.5"

EMBED_FCN: transformers.TransformersEmbeddingFunction = \
    get_registry().get("huggingface").create(name = EMBED_MODEL)


class TextChunk (LanceModel):
    """
Represents one chunk of text from a document.
    """
    uid: int
    url: str
    sent_id: int
    text: str = EMBED_FCN.SourceField()
    vector: Vector(EMBED_FCN.ndims()) = EMBED_FCN.VectorField(default = None)  # type: ignore # pylint: disable=E1136


@dataclass(order=False, frozen=False)
class Entity:  # pylint: disable=R0902
    """
Represents one entity in the graph.
    """
    loc: typing.Tuple[ int, int ]
    key: str
    text: str
    label: str
    chunk_id: int
    sent_id: int
    span: spacy.tokens.span.Span  # pylint: disable=I1101
    node: typing.Optional[ int ] = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
An adapted `TextRank` algorithm implementation based on `NetworkX` and `Polars`.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from collections import defaultdict
import itertools
import typing

import networkx as nx
import numpy as np
import pandas as pd
import polars as pl

from .elem import Entity, NodeKind, StrwVocab
from .opt import calc_quantile_bins, stripe_column, root_mean_square


def run_textrank (
    config: dict,
    lex_graph: nx.MultiDiGraph,
    ) -> pl.DataFrame:
    """
Run eigenvalue centrality (i.e., _Personalized PageRank_) to rank the entities.
    """
    # build a dataframe of node ranks and counts
    df_rank: pd.DataFrame = pd.DataFrame.from_dict([  # type: ignore
        {
            "node_id": node,
            "weight": rank,
            "count": lex_graph.nodes[node]["count"],
        }
        for node, rank in nx.pagerank(
                lex_graph,
                alpha = config["tr"]["tr_alpha"],
                weight = "count",
        ).items()
    ])

    # normalize by column and calculate quantiles
    df1: pd.DataFrame = df_rank[[ "count", "weight" ]].apply(lambda x: x / x.max(), axis = 0)
    bins: np.ndarray = calc_quantile_bins(len(df1.index))

    # stripe each columns
    df2: pd.DataFrame = pd.DataFrame([
        stripe_column(values, bins)  # type: ignore
        for _, values in df1.items()
    ]).T

    # renormalize the ranks
    df_rank["rank"] = df2.apply(root_mean_square, axis = 1)
    rank_col: np.ndarray = df_rank["rank"].to_numpy()
    rank_col /= sum(rank_col)
    df_rank["rank"] = rank_col

    # move the ranked weights back into the graph
    for _, row in df_rank.iterrows():
        node: int = row["node_id"]
        lex_graph.nodes[node]["rank"] = row["rank"]

    df: pl.DataFrame = pl.DataFrame([
        node_attr
        for node, node_attr in lex_graph.nodes(data = True)
        if node_attr["kind"] == NodeKind.ENTITY.value
    ]).sort(
        [ "rank" ],
        descending = True,
    )

    return df


def cooccur_entities (
    lex_graph: nx.MultiDiGraph,
    span_decoder: typing.Dict[ tuple, Entity ],
    ) -> None:
    """
Connect entities which co-occur within the same sentence.
    """
    ent_map: typing.Dict[ int, typing.Set[ int ]] = defaultdict(set)

    for ent in span_decoder.values():
        if ent.node is not None:
            ent_map[ent.sent_id].add(ent.node)

    for nodes in ent_map.values():
        for pair in itertools.combinations(list(nodes), 2):
            if not lex_graph.has_edge(*pair):
                lex_graph.add_edge(
                    pair[0],
                    pair[1],
                    key = StrwVocab.CO_OCCURS_WITH.value,
                    prob = 1.0,
                )
